"""
Copyright (C) 2019 - 2020 Julian Metzler

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

class BaseField:
    # This is needed so Form.get_fields() will know what to include
    _is_field = True
    
    def __init__(self, start_address = None, length = 1, descending = False,
                 text_align = 'left',
                 address_mapping = None, display_mapping = None,
                 x = 0, y = 0, module_width = 1, module_height = 1, home_pos = 0):
        """
        start_address: the address of the first module in this field
        length: How many modules make up this field
        descending: If using start_address, select descending addresses
        text_align: Alignment of the text (left, center, right)
        address_mapping: If modules have non-sequential addresses, the list of
                         addresses corresponding to the digits in this field
        display_mapping: Optional mapping of split-flap card numbers to
                         displayed text or symbols for all modules in this field
        x: Horizontal offset of the field in multiples of the smallest unit size
        y: Vertical offset of the field in multiples of the smallest unit size
        module_width: Width of the modules making up the field in multiples
                       of the smallest unit size
        module_height: Height of the modules making up the field in multiples
                       of the smallest unit size
        home_pos: ID of the home position. Should be 0, but is different in some cases
        """
        if start_address is None and address_mapping is None:
            raise AttributeError("Either start_address or address_mapping must be present")
        if type(length) is not int or length <= 0:
            raise ValueError("length must be a positive integer")
        if start_address is not None:
            if start_address not in range(256):
                raise ValueError("start_address must be an int between 0 and 255")
            if descending:
                if start_address - length < 0:
                    raise ValueError("Field is too long for given start address")
            else:
                if start_address + length > 256:
                    raise ValueError("Field is too long for given start address")
        if address_mapping is not None:
            if len(address_mapping) != length:
                raise ValueError("Length of address_mapping doesn't match field length")
        self.start_address = start_address
        self.length = length
        self.descending = descending
        if text_align not in ('left', 'center', 'right'):
            raise ValueError("text_align must be left, center or right")
        self.text_align = text_align
        if address_mapping is not None:
            self.address_mapping = address_mapping
        else:
            if self.descending:
                self.address_mapping = list(range(start_address, start_address-length, -1))
            else:
                self.address_mapping = list(range(start_address, start_address+length))
        self.display_mapping = display_mapping
        if display_mapping is not None:
            self.inverse_display_mapping = {v: k for k, v in display_mapping.items()}
        else:
            self.inverse_display_mapping = None
        self.x = x
        self.y = y
        self.module_width = module_width
        self.module_height = module_height
        self.home_pos = home_pos
        self.value = " " * self.length
        self.mirrors = []
    
    def set(self, value):
        self.value = value
    
    def get(self):
        return self.value
    
    def clear(self):
        self.value = " " * self.length
    
    def get_single_module_data(self, pos):
        raise NotImplementedError
    
    def get_module_data(self):
        module_data = []
        for i in range(self.length):
            module_data.append(self.get_single_module_data(i))
        return module_data
    
    def get_ascii_render_parameters(self):
        """
        Calculate the parameters needed to render the field as ASCII graphics
        """
        parameters = {
            'x': self.x * 2,
            'y': self.y * 2,
            'width': self.length * 2 * self.module_width + 1,
            'height': 2 * self.module_height + 1,
            'spacing': 2 * self.module_width,
            'text_spacing': 2 * self.module_width,
            'x_offset': self.module_width,
            'y_offset': self.module_height,
            'text_max_length': 2 * self.module_width - 1,
        }
        return parameters

    def add_mirror(self, field):
        """
        Add a field to the list of mirror fields
        """
        if field not in self.mirrors:
            self.mirrors.append(field)

    def remove_mirror(self, field):
        """
        Remove a field from the list of mirror fields
        """
        while field in self.mirrors:
            self.mirrors.remove(field)

    def update_mirrors(self):
        """
        Update all mirror fields of this field
        """
        for field in self.mirrors:
            if type(self.value) is list:
                field.value = self.value.copy()
            else:
                field.value = self.value


class MirrorField(BaseField):
    """
    This special field is set up so it mirrors an existing field.
    """
    def __init__(self, source_field, *args, **kwargs):
        if not isinstance(source_field, BaseField):
            raise ValueError("source_field must be an instance of a Field subclass")
        if 'start_address' not in kwargs:
            kwargs['start_address'] = source_field.start_address
        if 'length' not in kwargs:
            kwargs['length'] = source_field.length
        if 'descending' not in kwargs:
            kwargs['descending'] = source_field.descending
        if 'text_align' not in kwargs:
            kwargs['text_align'] = source_field.text_align
        if 'display_mapping' not in kwargs:
            kwargs['display_mapping'] = source_field.display_mapping
        if 'x' not in kwargs:
            kwargs['x'] = source_field.x
        if 'y' not in kwargs:
            kwargs['y'] = source_field.y
        if 'module_width' not in kwargs:
            kwargs['module_width'] = source_field.module_width
        if 'module_height' not in kwargs:
            kwargs['module_height'] = source_field.module_height
        if 'home_pos' not in kwargs:
            kwargs['home_pos'] = source_field.home_pos
        super().__init__(*args, **kwargs)
        self.source_field = source_field
        source_field.add_mirror(self)

    def set(self, value):
        pass

    def get(self):
        return self.source_field.get()

    def clear(self):
        pass

    def get_single_module_data(self, pos):
        addr, code = self.source_field.get_single_module_data(pos)
        x = self.x + pos * self.module_width
        return self.address_mapping[pos], code, x, self.y

    def get_module_data(self):
        module_data = []
        for i in range(self.length):
            module_data.append(self.get_single_module_data(i))
        return module_data

    def get_ascii_render_parameters(self):
        """
        Get the base parameters from the source field,
        but change the x and y values to allow for different placement
        """
        parameters = self.source_field.get_ascii_render_parameters()
        parameters.update({
            'x': self.x * 2,
            'y': self.y * 2,
        })
        return parameters


class TextField(BaseField):
    def __init__(self, *args, value = "", upper_only = True, **kwargs):
        super().__init__(*args, **kwargs)
        self.upper_only = upper_only
        self.set(value)
    
    def set(self, value):
        if type(value) is not str:
            raise ValueError("value must be str")
        if self.upper_only:
            value = value.upper()
        self.value = value[:self.length]
        if self.text_align == 'left':
            self.value = self.value.ljust(self.length)
        elif self.text_align == 'center':
            self.value = self.value.center(self.length)
        elif self.text_align == 'right':
            self.value = self.value.rjust(self.length)
        self.update_mirrors()
    
    def get_single_module_data(self, pos):
        """
        Returns the split-flap module address and code for the given position
        in the field with the current field value
        """
        if pos >= self.length:
            raise ValueError("pos must be inside field boundaries")
        addr = self.address_mapping[pos]
        char = self.value[pos]
        if self.display_mapping is not None:
            code = self.inverse_display_mapping.get(char, self.home_pos)
        else:
            code = ord(char.encode('iso-8859-1'))
        x = self.x + pos * self.module_width
        return addr, code, x, self.y


class CustomMapField(BaseField):
    def __init__(self, display_mapping, *args, value = [], **kwargs):
        super().__init__(*args, display_mapping=display_mapping, **kwargs)
        self.value = [""] * self.length
        self.set(value)

    def set(self, value):
        if type(value) not in (list, tuple):
            value = [value] * self.length
        value = value[:self.length] + [""] * (self.length - len(value))
        for i, module_value in enumerate(value):
            if module_value not in self.inverse_display_mapping:
                self.value[i] = ""
            else:
                self.value[i] = module_value
        self.update_mirrors()
    
    def clear(self):
        self.value = [""] * self.length
        self.update_mirrors()
    
    def get_single_module_data(self, pos):
        """
        Returns the split-flap module address and code for the given position
        in the field with the current field value
        """
        if pos >= self.length:
            raise ValueError("pos must be inside field boundaries")
        addr = self.address_mapping[pos]
        display_value = self.value[pos]
        code = self.inverse_display_mapping.get(display_value, self.home_pos)
        x = self.x + pos * self.module_width
        return addr, code, x, self.y
    
    def get_ascii_render_parameters(self):
        parameters = super().get_ascii_render_parameters()
        parameters['x_offset'] = 1
        parameters['text_spacing'] = 1
        return parameters


"""
Copyright (C) 2020 Julian Metzler

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import time


def calibrate_fbm_interactive(fbm, addr):
    """
    Interactive process for calibrating the flap count
    and home position on small FBM units
    """
    print("\nStep 0: Testing home position")
    print("Check if calibration is necessary by moving to the home position")
    print("and seeing if the unit hits it correctly.")
    input("Press Enter to start.")
    fbm.set_home()
    print("\nWait for the unit to stop rotating!")
    time.sleep(3)
    res = ""
    while res not in ("Y", "N"):
        res = input("Was the home position reached? [Y/N]: ").upper()
        if res == "Y":
            print("\nGreat! No need to calibrate.")
            return
    
    print("Step 1: Flap count (BR1) calibration")
    print("The unit will rotate once or twice and stop.")
    print("There is no need to do anything while it is rotating.")
    input("Press Enter to start.")
    fbm.start_calibration_br1()
    print("\nWait for the unit to stop rotating!")
    time.sleep(5)
    
    success = False
    back_to_step_2 = False
    while not success:
        exit_step_2 = False
        while not exit_step_2:
            print("\nStep 2: Home position (BR2) calibration")
            print("The unit will start rotating.")
            print("As soon as the desired home position is reached, hit Enter.")
            print("This might take a few tries.")
            print("The unit should complete one more rotation after you hit Enter!")
            print("If the unit stops rotating as soon as you hit Enter,")
            print("your timing was not correct. Try again in this case.")
            input("Press Enter to start and get ready to hit Enter again!")
            fbm.start_calibration_br2()
            input("Press Enter if the home position is reached!")
            fbm.stop_calibration()
            print("\nWait for the unit to stop rotating!")
            res = ""
            while res not in ("Y", "N"):
                res = input("Was the desired position reached? [Y/N]: ").upper()
                if res == "Y":
                    exit_step_2 = True
        
        print("\nStep 3: Testing home position")
        print("Test the calibration by moving to the home position again")
        print("and seeing if the unit hits it correctly.")
        input("Press Enter to start.")
        fbm.set_home()
        print("\nWait for the unit to stop rotating!")
        time.sleep(3)
        res = ""
        while res not in ("Y", "N"):
            res = input("Was the home position reached? [Y/N]: ").upper()
            if res == "N":
                success = False
                back_to_step_2 = True
                break
            elif res == "Y":
                back_to_step_2 = False
                break
        
        if back_to_step_2:
            continue
        
        exit_step_4 = False
        while not exit_step_4:
            print("\nStep 4: Testing characters")
            print("Test the calibration by entering letters")
            print("and seeing if the unit hits them correctly.")
            char = ""
            while len(char) != 1 or ord(char) not in range(128):
                char = input("Enter a character to test or nothing to finish: ").upper()
                if char == "":
                    exit_step_4 = True
                    success = True
                    break
                if len(char) > 1 or ord(char) not in range(128):
                    continue
                else:
                    fbm.set_code(addr, ord(char))
                    time.sleep(0.1)
                    fbm.set_all()
                    time.sleep(0.1)
                    res = ""
                    while res not in ("Y", "N"):
                        res = input("Was the desired character hit? [Y/N]: ").upper()
                        if res == "N":
                            exit_step_4 = True
                            success = False
    
    fbm.set_home()
    print("\nGreat! The unit is now calibrated.")



"""
Copyright (C) 2020 Julian Metzler

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import serial
import time


class OmegaRS485Controller:
    """
    Controls split-flap modules using Omega's RS485 protocol.
    Commonly found in Swiss (SBB CFF FFS) station displays.
    """

    def __init__(self, port, debug = False, exclusive = False):
        self.debug = debug
        self.port = serial.Serial(port, baudrate=19200, timeout=1.0, exclusive=exclusive)

    def prepare_message(self, address, command, value):
        message = [0xFF, command, address]
        if value is not None:
            if type(value) in (tuple, list):
                message.extend(value)
            else:
                message.append(value)
        return message

    def init_communication(self):
        """
        Initialize communication by asserting a break condition
        on the serial Tx line for a certain time
        """
        self.port.break_condition = True
        time.sleep(0.05)
        self.port.break_condition = False

    def send_raw_message(self, message):
        if self.debug:
            print(" ".join((format(x, "02X") for x in message)))
        self.init_communication()
        self.port.write(message)

    def read_response(self, length):
        return self.port.read(length)

    def send_command(self, address, command, value = None):
        message = self.prepare_message(address, command, value)
        self.send_raw_message(message)

    def set_home(self, address):
        self.send_command(address, 0xC5)

    def set_position(self, address, position):
        self.send_command(address, 0xC0, position)

    def set_address(self, address, new_address):
        self.send_command(address, 0xCE, new_address)

    def read_position(self, address):
        self.send_command(address, 0xD0)
        return self.read_response(4)

    def read_serial_number(self, address):
        self.send_command(address, 0xDF)
        return self.read_response(1)

    def d_set_module_data(self, module_data):
        # Compatibility function for SplitFlapDisplay class
        
        # Turn module data into blocks of contiguous addresses
        items = sorted(module_data, key=lambda i:i[0])
        last_addr = None
        start_addr = None
        pos_block = []
        for i, (addr, pos) in enumerate(items):
            if (last_addr is not None and addr - last_addr > 1):
                self.set_position(start_addr, pos_block)
                time.sleep(0.05)
                pos_block = []
            if pos_block == []:
                start_addr = addr
            pos_block.append(pos)
            if i == len(items) - 1:
                self.set_position(start_addr, pos_block)
                time.sleep(0.05)
                pos_block = []
            last_addr = addr

    def d_update(self):
        # Compatibility function for SplitFlapDisplay class
        pass



"""
Copyright (C) 2019 - 2020 Julian Metzler

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import serial
import time

from .exceptions import CommunicationError, NACKError, BusyError


class Krone9000RS485Controller:
    """
    Basic control scheme for the internal RS485 bus
    in the KRONE 9000 system.
    """
    
    FLAG_ACK = 0x80
    
    CTRL_READ = 0x01
    CTRL_WRITE_SINGLE = 0x02
    CTRL_WRITE_BLOCK = 0x04
    
    STATUS_ACK = 0x06
    STATUS_BUSY = 0x10
    STATUS_NACK = 0x15
    
    MAX_CHUNK_SIZE = 128
    RETRY_COUNT = 10
    RETRY_INTERVAL = 0.2

    def __init__(self, port, address, timeout = 1.0, debug = False, exclusive = False):
        self.address = address
        self.debug = debug
        if isinstance(port, serial.Serial):
            self.port = port
        else:
            self.port = serial.Serial(port, baudrate=19200, timeout=timeout, exclusive=exclusive)
    
    @staticmethod
    def _chunks(lst, n):
        """Yield successive n-sized chunks from lst."""
        for i in range(0, len(lst), n):
            yield lst[i:i + n]
    
    def debug_print(self, data, receive = False):
        if self.debug:
            print(f"{self.BOARD_NAME:<4} {self.address} {'RX' if receive else 'TX'}: " + " ".join((format(x, "02X") for x in data)))
    
    def make_checksum(self, payload):
        checksum = 0x00
        for byte in payload:
            checksum ^= byte
        return checksum
    
    def check_checksum(self, data):
        checksum = self.make_checksum(data[2:-1])
        return checksum == data[-1]
            
    def send_command(self, command, parameters = None, response = False, ack = True, block = False):
        def _chunks(lst, n):
            # Yield successive n-sized chunks from lst.
            for i in range(0, len(lst), n):
                yield lst[i:i + n]
        
        data = [command]
        if parameters is not None:
            if type(parameters) in (list, tuple):
                data.extend(parameters)
            else:
                data.append(parameters)
        
        if len(data) > self.MAX_CHUNK_SIZE and block == False:
            raise CommunicationError("Data too long for single-block transfer. Use multi-block mode!")
        
        if response:
            control = self.CTRL_READ
        else:
            if block:
                control = self.CTRL_WRITE_BLOCK
            else:
                control = self.CTRL_WRITE_SINGLE
        
        if ack:
            control |= self.FLAG_ACK
        
        for chunk_id, chunk in enumerate(_chunks(data, self.MAX_CHUNK_SIZE)):
            for retry in range(self.RETRY_COUNT):
                payload = [self.BOARD_ID, self.address, 0x00, control, chunk_id + 1] + chunk + [0x00]
                length = len(payload)
                payload[2] = length
                checksum = self.make_checksum(payload)
                cmd_bytes = [0xFF, 0xFF] + payload + [checksum]
                
                # Debug output if enabled
                self.debug_print(cmd_bytes)
                
                # Send it
                self.port.write(bytearray(cmd_bytes))
                
                # Check status
                if not response:
                    if control & self.FLAG_ACK:
                        try:
                            self.check_status()
                        except BusyError:
                            if retry >= self.RETRY_COUNT - 1:
                                raise
                            else:
                                time.sleep(self.RETRY_INTERVAL)
                        else:
                            break
                else:
                    break

        # Read response
        if response:
            return self.read_response()
        else:
            return None
    
    def check_status(self):
        status = bytearray(self.port.read(1))[0]
        self.debug_print([status], receive=True)
        if status == self.STATUS_ACK:
            return True
        elif status == self.STATUS_BUSY:
            raise BusyError()
        elif status == self.STATUS_NACK:
            raise NACKError()
        else:
            raise CommunicationError(f"Unknown status byte {status:02X}")
    
    def read_response(self):
        data = bytearray()
        header = bytearray(self.port.read(4))
        if len(header) != 4:
            if len(header) == 1:
                if header[0] == self.STATUS_BUSY:
                    raise BusyError()
                if header[0] == self.STATUS_NACK:
                    raise NACKError()
            header_fmt = " ".join((format(x, "02X") for x in header))
            raise CommunicationError(f"Incomplete header: {header_fmt}")
        if header[0] != 0xFF or header[1] != 0xFF:
            raise CommunicationError(f"Invalid start sequence {header[0]:02X} {header[1]:02X}")
        if header[2] != self.BOARD_ID:
            raise CommunicationError(f"Invalid board ID {header[2]:02X}")
        if header[3] != self.address:
            raise CommunicationError(f"Invalid address {header[3]:02X}")
        data.extend(header)
        
        length = bytearray(self.port.read(1))[0]
        payload = bytearray(self.port.read(length - 3))
        checksum = bytearray(self.port.read(1))[0]
        
        data.append(length)
        data.extend(payload)
        data.append(checksum)
        
        self.debug_print(data, receive=True)
        if not self.check_checksum(data):
            raise CommunicationError("Checksum mismatch")
        return payload[2:-1]
    
    def send_heartbeat(self):
        cmd_bytes = [0xFF, 0xFF, self.BOARD_ID, self.address, 0x00]
        self.debug_print(cmd_bytes)
        self.port.write(bytearray(cmd_bytes))


"""
Copyright 2020 Julian Metzler

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import serial
import time


class Krone9000FBM:
    """
    Controls one or several FBM (Fallblattmodul)
    (split flap module) boards.
    """
    
    CMD_SET_ALL = 0b0001
    CMD_SET_HOME = 0b0010
    CMD_RESET = 0b0011
    CMD_READ_STATUS = 0b0100
    CMD_READ_CODE = 0b0101
    CMD_LOCK = 0b0110
    CMD_UNLOCK = 0b0111
    CMD_SET_CODE = 0b1000
    CMD_START_CALIBRATION_BR2 = 0b1001
    CMD_START_CALIBRATION_BR1 = 0b1010
    CMD_STOP_CALIBRATION = 0b1011
    CMD_GET_CALIBRATION_VALUES = 0b1100
    CMD_SET_TABLE = 0b1101
    CMD_DELETE_TABLE = 0b1110
    
    def _get_fbm_status(self, stat):
        # Return human-readable error strings
        # based on FBM error bits
        if stat & 0x08:
            lut = {
                0b1000: "comm_error",
                0b1001: "start_missing",
                0b1010: "unknown_char",
                0b1011: "external_rotation",
                0b1100: "rotation_timeout",
                0b1101: "fbm_missing",
                0b1111: "rotating"
            }
            return [lut.get(stat & 0x0f, "")]
        else:
            errors = []
            if stat & 0x04:
                errors.append("no_ac")
            if stat & 0x02:
                errors.append("no_flap_imps")
            if stat & 0x01:
                errors.append("no_home_imp")
            return errors

    def __init__(self, port, debug = False, exclusive = False):
        self.debug = debug
        self.port = serial.Serial(port, baudrate=4800, parity=serial.PARITY_EVEN, timeout=2.0, exclusive=exclusive)
    
    def send_command(self, command, address = None, code = None, position = None, num_response_bytes = 0):
        # Build base command byte
        cmd_bytes = []
        cmd_base = 0b10010000

        # Address expansion bit
        if address is not None and address > 127:
            cmd_base |= 0b01000000
        
        # Code expansion bit
        if code is not None and (code > 127 or (position is not None and position > 127)):
            cmd_base |= 0b00100000
        
        # Command bits
        cmd_base |= command
        
        # Build the data to be sent
        cmd_bytes.append(cmd_base)
        if address is not None:
            cmd_bytes.append(address & 0b01111111)
        if code is not None:
            cmd_bytes.append(code & 0b01111111)
        if position is not None:
            cmd_bytes.append(position & 0b01111111)
        
        if self.debug:
            print(" ".join((format(x, "#010b") for x in cmd_bytes)))
            print(" ".join((format(x, "02X") for x in cmd_bytes)))
        
        # Send it
        self.port.write(bytearray(cmd_bytes))

        # Read response
        if num_response_bytes > 0:
            return self.port.read(num_response_bytes)
        else:
            return None
    
    def set_all(self):
        return self.send_command(self.CMD_SET_ALL)
    
    def set_home(self):
        return self.send_command(self.CMD_SET_HOME)
    
    def reset(self):
        return self.send_command(self.CMD_RESET)
    
    def read_status(self, address):
        return self.send_command(self.CMD_READ_STATUS, address, num_response_bytes=1)
    
    def read_code(self, address):
        return self.send_command(self.CMD_READ_CODE, address, num_response_bytes=1)
    
    def lock(self, address):
        return self.send_command(self.CMD_LOCK, address)
    
    def unlock(self, address):
        return self.send_command(self.CMD_UNLOCK, address)
    
    def set_code(self, address, code):
        return self.send_command(self.CMD_SET_CODE, address, code)
    
    def start_calibration_br2(self):
        return self.send_command(self.CMD_START_CALIBRATION_BR2)
    
    def start_calibration_br1(self):
        return self.send_command(self.CMD_START_CALIBRATION_BR1)
    
    def stop_calibration(self):
        return self.send_command(self.CMD_STOP_CALIBRATION)
    
    def get_calibration_values(self, address):
        return self.send_command(self.CMD_GET_CALIBRATION_VALUES, address, num_response_bytes=1)
    
    def set_table(self, address, code, position):
        return self.send_command(self.CMD_SET_TABLE, address, position, code, num_response_bytes=1)
    
    def delete_table(self, address):
        return self.send_command(self.CMD_DELETE_TABLE, address, num_response_bytes=1)
    
    def set_text(self, text, start_address, length = None, descending = False):
        if length is not None:
            text = text[:length].ljust(length)
        for i, char in enumerate(text):
            address = start_address - i if descending else start_address + i
            self.set_code(address, ord(char.encode('iso-8859-1')))
    
    def get_status(self, addr):
        return self._get_fbm_status(self.read_status(addr)[0])
    
    def d_set_module_data(self, module_data):
        # Compatibility function for SplitFlapDisplay class
        for addr, code in module_data:
            self.set_code(addr, code)
    
    def d_update(self):
        # Compatibility function for SplitFlapDisplay class
        self.set_all()


"""
Copyright (C) 2019 - 2020 Julian Metzler

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import serial

from itertools import groupby

from .k9000_rs485 import Krone9000RS485Controller
from .exceptions import CommunicationError


class Krone9000FBK(Krone9000RS485Controller):
    """
    Controls a FBK (Fallblatt-Buskopfkarte)
    (split-flap group control) board.
    """
    
    BOARD_ID = 0xB2
    BOARD_NAME = "FBK"
    
    FLAG_START_IMMEDIATELY = 0x80
    FLAG_FBM_COMMAND = 0x40
    FLAG_ENABLE_COMPRESSION = 0x20
    
    CMD_GET_FBK_STATUS = 0x01
    CMD_GET_LINE_INIT_DATA = 0x02
    CMD_GET_LINE_DATA = 0x03
    CMD_SET_BLINKER = 0x0B
    CMD_SET_FBM_VALUE_TABLE = 0x0E
    CMD_LOCK_FBK = 0xC6
    CMD_UNLOCK_FBK = 0xC7
    
    CMD_FBM_START = 0x00 | FLAG_START_IMMEDIATELY
    CMD_GET_FBM_STATUS = 0x03 | FLAG_FBM_COMMAND
    CMD_GET_FBM_CONTENT = 0x04 | FLAG_FBM_COMMAND
    CMD_CLEAR_FBM = 0x05 | FLAG_FBM_COMMAND
    CMD_LOCK_FBM = 0x06 | FLAG_FBM_COMMAND | FLAG_START_IMMEDIATELY
    CMD_UNLOCK_FBM = 0x07 | FLAG_FBM_COMMAND | FLAG_START_IMMEDIATELY
    CMD_SET_FBM_CODES_SEQ = 0x08 | FLAG_FBM_COMMAND
    CMD_SET_FBM_CODES_ADDR = 0x09 | FLAG_FBM_COMMAND
    
    def _get_fbm_status(self, stat):
        # Return human-readable error strings
        # based on FBM error bits
        if stat & 0x08:
            lut = {
                0b1000: "comm_error",
                0b1001: "start_missing",
                0b1010: "unknown_char",
                0b1011: "external_rotation",
                0b1100: "rotation_timeout",
                0b1101: "fbm_missing",
                0b1111: "rotating"
            }
            return [lut.get(stat & 0x0f, "")]
        else:
            errors = []
            if stat & 0x04:
                errors.append("no_ac")
            if stat & 0x02:
                errors.append("no_flap_imps")
            if stat & 0x01:
                errors.append("no_home_imp")
            return errors
    
    def _rle(self, data):
        # Simple run-length encoding
        rle_data = []
        for k,i in groupby(data):
            run = list(i)
            if(len(run) > 2):
                while len(run) > 128:
                    rle_data.extend([(len(run) - 1) | 0x80, k])
                    run = run[128:]
                rle_data.extend([(len(run) - 1) | 0x80, k])
            else:
                rle_data.extend(run)
        return rle_data
    
    def get_status(self):
        # Get the status of the FBK board
        payload = self.send_command(self.CMD_GET_FBK_STATUS, response=True)
        stat = payload[1]
        status = {
            'comm_err': bool(stat & 0x40),
            'reset': bool(stat & 0x20),
            'locked': bool(stat & 0x10),
            'fbm_err': bool(stat & 0x08),
            'blinker_err': bool(stat & 0x04),
            'fbm_start_missing': bool(stat & 0x02),
            'sw_ver': f"{payload[2]}.{payload[3]}"
        }
        return status
    
    def get_fbm_ids(self):
        # Get a list of all connected FBM IDs
        payload = self.send_command(self.CMD_GET_LINE_INIT_DATA, response=True)
        module_data = payload[1:]
        modules_present = []
        addr = 0
        for byte in module_data:
            for bit in range(7):
                if byte & (1 << bit):
                    modules_present.append(addr)
                addr += 1
                if addr > 255:
                    break
        return modules_present
    
    def get_fbm_statuses(self):
        # True means okay, False means FBM error or not present
        # (see get_fbm_ids to get a list of present FBMs)
        payload = self.send_command(self.CMD_GET_LINE_DATA, response=True)
        module_data = payload[1:]
        module_statuses = {}
        addr = 0
        for byte in module_data:
            for bit in range(7):
                module_statuses[addr] = bool(byte & (1 << bit))
                addr += 1
                if addr > 255:
                    break
        return module_statuses
    
    def set_blinker(self, state):
        # Set the blinker associated with this FBK on or off
        return self.send_command(self.CMD_SET_BLINKER, 0x31 if state else 0x30)
    
    def set_fbm_value_table(self, addr, table):
        # Set the internal mapping of character code to flap position
        # on the selected FBM
        # table is a list of characters to be mapped to flaps,
        # starting at flap 0
        parameters = [addr] + table
        self.send_command(self.CMD_SET_FBM_VALUE_TABLE, parameters)
        # Special case here: The FBK sends the number of flap codes transferred
        # as a single byte after it has finished sending the data to the FBM.
        # As this is the only case in which this sort of response occurs,
        # we are just handling it manually here.
        num_xferred = b""
        tries = 0
        # Try for roughly five seconds (exact value depends on the port timeout)
        while num_xferred == b"":
            if tries >= 5.0 / self.port.timeout:
                raise CommunicationError("Timeout while waiting for result of FBM value table transfer")
            num_xferred = self.port.read(1)
            tries += 1
        self.debug_print(bytearray(num_xferred), receive=True)
        # Return the number of transferred flap codes
        return ord(num_xferred)
    
    def lock(self):
        # Lock the entire FBK
        return self.send_command(self.CMD_LOCK_FBK)
    
    def unlock(self):
        # Unlock the entire FBK
        return self.send_command(self.CMD_UNLOCK_FBK)
    
    def start_fbm(self):
        return self.send_command(self.CMD_FBM_START)
    
    def get_detailed_fbm_statuses(self, addrs):
        # Get detailed FBM status information for up to 10 FBMs.
        # addrs is a list of FBM IDs to be queried
        payload = self.send_command(self.CMD_GET_FBM_STATUS, addrs, response=True)
        module_statuses = {}
        data = payload[1:]
        for i in range(0, len(data), 2):
            addr = data[i]
            stat = data[i + 1]
            module_statuses[addr] = {
                'home_pos': bool(stat & 0x40),
                'reset': bool(stat & 0x20),
                'locked': bool(stat & 0x10),
                'status': self._get_fbm_status(stat & 0x0f)
            }
        return module_statuses
    
    def get_all_detailed_fbm_statuses(self):
        # Automatically read the list of connected FBM IDs
        # and query all of them for detailed status information
        module_statuses = {}
        for addrs in self._chunks(self.get_fbm_ids(), 10):
            module_statuses.update(self.get_detailed_fbm_statuses(addrs))
        return module_statuses
    
    def get_fbm_contents(self, addrs):
        # Get the currently displayed character for up to 10 FBMs.
        # addrs is a list of FBM IDs to be queried
        payload = self.send_command(self.CMD_GET_FBM_CONTENT, addrs, response=True)
        module_contents = {}
        data = payload[1:]
        for i in range(0, len(data), 2):
            addr = data[i]
            char = data[i + 1]
            module_contents[addr] = chr(char)
        return module_contents
    
    def get_all_fbm_contents(self):
        # Automatically read the list of connected FBM IDs
        # and query all of them for their content
        module_contents = {}
        for addrs in self._chunks(self.get_fbm_ids(), 10):
            module_contents.update(self.get_fbm_contents(addrs))
        return module_contents
    
    def clear_fbm(self, addrs = None, immediate = True):
        # Clear all or the selected FBMs
        cmd = self.CMD_CLEAR_FBM
        if immediate:
            cmd |= self.FLAG_START_IMMEDIATELY
        return self.send_command(cmd, addrs)
    
    def lock_fbm(self, addrs = None):
        # Lock all or the selected FBMs
        return self.send_command(self.CMD_LOCK_FBM, addrs)
    
    def unlock_fbm(self, addrs = None):
        # Unlock all or the selected FBMs
        return self.send_command(self.CMD_UNLOCK_FBM, addrs)
    
    def set_fbm_codes_seq(self, codes, immediate = False, compress = False):
        # Set FBM character codes to be displayed
        # sequentially (without explicit addressing)
        cmd = self.CMD_SET_FBM_CODES_SEQ
        if immediate:
            cmd |= self.FLAG_START_IMMEDIATELY
        if compress:
            cmd |= self.FLAG_ENABLE_COMPRESSION
            codes = self._rle(codes)
        return self.send_command(cmd, codes)
    
    def set_fbm_codes_addr(self, codes, immediate = False):
        # Set FBM character codes to be displayed
        # with explicit addressing
        cmd = self.CMD_SET_FBM_CODES_ADDR
        if immediate:
            cmd |= self.FLAG_START_IMMEDIATELY
        return self.send_command(cmd, codes)
    
    def d_set_module_data(self, module_data):
        # Compatibility function for SplitFlapDisplay class
        for chunk in self._chunks(module_data, 50):
            self.set_fbm_codes_addr([i for s in chunk for i in s])
    
    def d_update(self):
        # Compatibility function for SplitFlapDisplay class
        self.start_fbm()



"""
Copyright (C) 2016 - 2020 Julian Metzler

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import math

class IBISProtocol:
    """
    All the logic related to the IBIS protocol
    """
    
    def __init__(self, debug = False):
        """
        debug:
        Whether to print the sent and received telegrams
        """
        
        self.debug = debug
        
        # Simple telegram definitions
        self.DS001 = self._tg("l{:>03}")        # Line number, 1-4 digits
        self.DS001neu = self._tg("q{:0>4}")     # Line number, alphanumeric, 1-4 chars
        self.DS001a = self._tg("qE{:02d}")      # Line number symbol ID, 1-2 digit
        self.DS001b = self._tg("lF{:05d}")      # Radio
        self.DS001c = self._tg("lP{:03d}")      # Line tape reel position ID, 1-3 digit
        self.DS001d = self._tg("lC{:0>4}")      # Line number, alphanumeric, 1-4 chars
        self.DS001e = self._tg("lC{:0>8}")      # Line number, alphanumeric, 1-8 chars
        self.DS001f = self._tg("lB{:0>7}")      # Line number, alphanumeric, 1-7 chars
        self.DS002 = self._tg("k{:02d}")        # Course number, 1-2 digits
        self.DS002a = self._tg("k{:05d}")       # Train number, 1-5 digits
        self.DS003 = self._tg("z{:03d}")        # Destination text ID, 1-3 digit
        self.DS003b = self._tg("zR{:03d}")      # Destination ID for IMU, 1-3 digits
        self.DS003d = self._tg("zN{:03d}")      # Route number, 1-3 digits
        self.DS003e = self._tg("zP{:03d}")      # Destination tape reel position ID, 1-3 digit
        self.DS003f = self._tg("zN{:06d}")      # Route number, 1-6 digits
        self.DS003g = self._tg("zL{:04d}")      # Line number, 1-4 digits
        self.DS004 = self._tg("e{:06d}")        # Ticket validator attributes, 6 digits
        self.DS004a = self._tg("eA{:04d}")      # Additional ticket validator attributes, 4 digits
        self.DS004b = self._tg("eH{:07d}")      # Ticket validator stop number, 1-7 digits
        self.DS005 = self._tg("u{:04d}")        # Time, HHMM
        self.DS006 = self._tg("d{:05d}")        # Date, DDMMY
        self.DS007 = self._tg("w{:01d}")        # Train length, 1 digit
        self.DS009 = self._tg("v{: <16}")       # Next stop text, 16 characters
        self.DS009a = self._tg("v{: <20}")      # Next stop text, 20 characters
        self.DS009b = self._tg("v{: <24}")      # Next stop text, 24 characters
        self.DS010 = self._tg("x{:04d}")        # Line progress display stop ID, 1-4 digits
        self.DS010a = self._tg("xH{:04d}")      # Line progress display stop ID, 1-4 digits
        self.DS010b = self._tg("xI{:02d}")      # Line progress display stop ID, 1-2 digits
        self.DS010d = self._tg("xJ{:04d}")      # Year, YYYY
        self.DS010e = self._tg("xV{}{:03d}")    # Delay, +/-, 1-3 digits
    
    def _send(self, telegram):
        """
        Actually send the telegram.
        This varies depending on implementation and needs to be overridden
        """
        pass
    
    def _receive(self, length):
        """
        Actually receive data.
        This varies depending on implementation and needs to be overridden
        """
        pass

    def _printable(self, char):
        """
        Replace non-printable chars with printable variants.
        """
        if char in range(0, 32):
            return "<{:02X}>".format(char)
        else:
            return chr(char)
    
    def debug_telegram(self, telegram, receive = False):
        """
        Print a telegram to standard output if the debug flag is set.
        
        telegram:
        The telegram to print
        
        receive:
        Whether to print the telegram as sent or received
        """
        
        if self.debug:
            action = "Received" if receive else "Sending"
            telegram_debug = "\n".join("{:02X} {}".format(
                byte, self._printable(byte)) for byte in telegram)
            print("{} telegram:\n".format(action))
            print(telegram_debug)
    
    def process_special_characters(self, telegram):
        """
        Replace IBIS special characters and strip unsupported characters.
        
        telegram:
        The telegram as a string
        
        Returns:
        The processed telegram
        """
        
        telegram = telegram.replace("ä", "{")
        telegram = telegram.replace("ö", "|")
        telegram = telegram.replace("ü", "}")
        telegram = telegram.replace("ß", "~")
        telegram = telegram.replace("Ä", "[")
        telegram = telegram.replace("Ö", "\\")
        telegram = telegram.replace("Ü", "]")
        telegram = telegram.encode('ascii', errors = 'replace')
        return telegram
    
    def wrap_telegram(self, telegram):
        """
        Append the checksum and the end byte to the given telegram.
        
        telegram:
        The telegram (as a bytearray) to wrap
        
        Returns:
        The wrapped telegram (as a bytearray)
        """
        
        telegram.append(0x0D)
        checksum = 0x7F
        for byte in telegram:
            checksum ^= byte
        telegram.append(checksum)
        return telegram
    
    def send_telegram(self, telegram, reply_length = 0):
        """
        Send an arbitrary telegram. Checksum and end byte will be added.
        
        telegram:
        The telegram to send
        
        reply_length:
        How many bytes to expect as a reply
        
        Returns:
        The received telegram OR None
        
        TODO: Actually check the checksum
        """
        
        if type(telegram) is str:
            telegram = self.process_special_characters(telegram)
            telegram = bytearray(telegram)
        elif type(telegram) is bytes:
            telegram = bytearray(telegram)
        
        telegram = self.wrap_telegram(telegram)
        self.debug_telegram(telegram)
        self._send(telegram)
        if reply_length:
            reply = self._receive(reply_length + 2)
            self.debug_telegram(reply, receive = True)
            reply = reply[:-2]
            return reply.decode('latin1')
    
    def vdv_hex(self, value):
        """
        Convert a numerical value into the VDV hexadecimal representation
        
        value:
        The value to convert (0 to 255) OR a VDV Hex value to convert back
        
        Returns:
        The VDV Hex value OR the integer for the VDV Hex value
        """
        
        vdvhex = "0123456789:;<=>?"
        if type(value) is int:
            assert 0 <= value <= 255
            if value > 15:
                high_nibble = value >> 4
                low_nibble = value % 16
                return vdvhex[high_nibble] + vdvhex[low_nibble]
            else:
                return vdvhex[value]
        else:
            assert 1 <= len(value) <= 2
            if len(value) == 2:
                high_nibble = vdvhex.index(value[0])
                low_nibble = vdvhex.index(value[1])
                return high_nibble << 4 + low_nibble
            else:
                return vdvhex.index(value)
    
    def _tg(self, fmt, reply_length = 0):
        """
        Wrapper for simple telegrams with just variables
        
        fmt:
        The format string for the telegram
        
        reply_length:
        As in send_telegram
        """
        
        def _send(*args):
            return self.send_telegram(fmt.format(*args),
                reply_length = reply_length)
        
        return _send
    
    def DS003a(self, text):
        """
        Destination text
        
        text:
        The destination text
        """
        
        num_blocks = math.ceil(len(text) / 16)
        return self.send_telegram("zA{}{}"
            .format(self.vdv_hex(num_blocks), text.ljust(num_blocks*16)))
    
    def DS003aUESTRA(self, front_text, side_text = "", line_text = "",
        display_line_text_front = True, display_line_text_side = True,
        display_interval_front = 3.0, display_interval_side = 3.0,
        text_align_front = "M", text_align_side = "M",
        bold_text_front = True, bold_text_side = True):
        """
        Destination text for ÜSTRA displays
        
        front_text:
        The destination text for the front display
        (string or array of up to 4 strings to be displayed sequentially)

        side_text:
        The destination text for the side display
        (string or array of up to 4 strings to be displayed sequentially)

        line_text:
        The line number (string, alphanumerical)

        display_line_text_front:
        Whether to display line number on the front display
        (bool or array of up to 4 bools to set line number display
        per text block)

        display_line_text_side:
        Whether to display line number on the side display
        (bool or array of up to 4 bools to set line number display
        per text block)
        
        display_interval_front:
        Interval for switching between sequential text blocks in seconds
        on the front display (1.0 ... 8.5 seconds, in 0.5 second steps)
        
        display_interval_side:
        Interval for switching between sequential text blocks in seconds
        on the side display (1.0 ... 8.5 seconds, in 0.5 second steps)
        
        text_align_front:
        Text align for front text. (L=Left, M=Middle, R=Right)
        
        text_align_side:
        Text align for side text. (L=Left, M=Middle, R=Right)
        
        bold_text_front:
        Whether text should be bold or thin on the front display.
        (bool or array of bools for block 1 line 1, block 1 line 2,
        block 2 line 1, block 2 line 2, etc. up to block 4 line 2)
        
        bold_text_side:
        Whether text should be bold or thin on the side display.
        (bool or array of bools for block 1 line 1, block 1 line 2,
        block 2 line 1, block 2 line 2, etc. up to block 4 line 2)
        """

        def _insert_case_switch_control_chars(text):
            """
            Insert the case-switching control character 0x06 in the given text
            to switch between upper and lower case
            """
            if not text:
                return ""
            ret_text = ""
            is_lower = False # Keep track of the current case
            for char in text:
                if not is_lower and ord(char) in range(0x60, 0x80):
                    ret_text += chr(0x06)
                    is_lower = True
                if is_lower and ord(char) in range(0x40, 0x60):
                    ret_text += chr(0x06)
                    is_lower = False
                if char == "\n":
                    is_lower = False
                ret_text += char.upper()
            if is_lower:
                # Make sure we are back to upper
                ret_text += chr(0x06)
            return ret_text
        
        def _array_to_byte(array):
            """
            Convert a bool array to a byte with the corresponding bits set
            """
            byte = 0x00
            for i, val in enumerate(array):
                byte |= (int(val) << i)
            return byte
        
        if type(front_text) in (bytes, str):
            front_text = [front_text]
        if type(side_text) in (bytes, str):
            side_text = [side_text]
        if type(display_line_text_front) is bool:
            display_line_text_front = [display_line_text_front] * 4
        if type(display_line_text_side) is bool:
            display_line_text_side = [display_line_text_side] * 4
        if type(bold_text_front) is bool:
            bold_text_front = [bold_text_front] * 8
        if type(bold_text_side) is bool:
            bold_text_side = [bold_text_side] * 8
        
        display_interval_front = max(1.0, display_interval_front)
        display_interval_front = min(display_interval_front, 8.5)
        display_interval_front = round(display_interval_front * 2) - 2 # 0 to 15
        
        display_interval_side = max(1.0, display_interval_side)
        display_interval_side = min(display_interval_side, 8.5)
        display_interval_side = round(display_interval_side * 2) - 2 # 0 to 15
        
        front_text_lines = [a + [""] * (2-len(a)) for a in [_insert_case_switch_control_chars(t).splitlines() for t in front_text]]
        side_text_lines = [a + [""] * (2-len(a)) for a in [_insert_case_switch_control_chars(t).splitlines() for t in side_text]]
        line_text = _insert_case_switch_control_chars(line_text)
        
        data = ""
        for lines in front_text_lines:
            data += "\n.W{}\n{}\n".format(lines[0], lines[1])
        for lines in side_text_lines:
            data += "\n.X{}\n{}\n".format(lines[0], lines[1])
        data += "\n.Y{}\n".format(line_text)
        data += "\n.C"
        data += chr(0x30 + _array_to_byte(display_line_text_front))
        data += chr(0x30 + _array_to_byte(display_line_text_side))
        data += chr(0x30 + display_interval_front)
        data += chr(0x30 + display_interval_side)
        data += text_align_front
        data += text_align_side
        data += "M"
        data += chr(0x20 + _array_to_byte(bold_text_front[0:6]))
        data += chr(0x20 + _array_to_byte(bold_text_front[6:8] + bold_text_side[0:4]))
        data += chr(0x20 + _array_to_byte(bold_text_side[4:8]))
        
        num_blocks = math.ceil(len(data) / 4)
        return self.send_telegram("zA{:>02}{}"
            .format(self.vdv_hex(num_blocks), data.ljust(num_blocks*4)))
    
    def DS003c(self, text):
        """
        Next stop text
        
        text:
        The next stop text
        """
        
        num_blocks = math.ceil(len(text) / 4)
        return self.send_telegram("zI{}{}"
            .format(self.vdv_hex(num_blocks), text.ljust(num_blocks*4)))
    
    def DS004c(self, text):
        """
        Stop text for ticket validators / ticket vending machines
        
        text:
        The stop text
        """
        
        num_blocks = math.ceil(len(text) / 4)
        return self.send_telegram("eT{}{}"
            .format(self.vdv_hex(num_blocks), text.ljust(num_blocks*4)))
    
    def DS010c(self, stop_id):
        """
        Next stop ID for line progress display
        
        stop_id:
        The ID of the next stop, 1-2 digits
        """
        
        stop_id_high_nibble = stop_id >> 4
        stop_id_low_nibble = stop_id & 0x0F
        return self.send_telegram("xZ{}{}"
            .format(self.vdv_hex(stop_id_high_nibble),
                self.vdv_hex(stop_id_low_nibble)))
    
    def DS010f(self, stop_id, change_text):
        """
        Connection information for line progress display
        
        stop_id:
        The ID of the next stop, 1-2 digits
        
        change_text:
        Connection information
        """
        
        stop_id_high_nibble = stop_id >> 4
        stop_id_low_nibble = stop_id & 0x0F
        length_high_nibble = len(change_text) >> 4
        length_low_nibble = len(change_text) & 0x0F
        return self.send_telegram("xU{}{}{}{}{}"
            .format(self.vdv_hex(stop_id_high_nibble),
                self.vdv_hex(stop_id_low_nibble),
                self.vdv_hex(length_high_nibble),
                self.vdv_hex(length_low_nibble),
                change_text))
    
    def DS020(self, address):
        """
        Display status query
        Reply: DS120
        
        address:
        The address of the display
        """
        
        return self.parse_DS120(self.send_telegram("a{}"
            .format(self.vdv_hex(address)), reply_length = 2))
    
    def parse_DS120(self, telegram):
        if not telegram:
            return None
        status = telegram[1]
        statuses = {
            '0': 'ok',
            '1': 'displaying',
            '2': 'searching',
            '3': 'error',
            '6': 'input_implausible'
        }
        reply = {
            'status': statuses.get(status, status)
        }
        return reply
    
    def DS201(self, address):
        """
        Display version query
        Reply: DS1201
        
        address:
        The address of the display
        """
        
        return self.parse_DS1201(self.send_telegram("aV{}"
            .format(self.vdv_hex(address)), reply_length = 8))
    
    def parse_DS1201(self, telegram):
        if not telegram:
            return None
        version = telegram[2:]
        reply = {
            'version': version
        }
        return reply
    
    def DS021(self, address, text):
        """
        Destination text
        
        address:
        The address of the display
        
        text:
        The destination text
        """
        
        num_blocks = math.ceil(len(text) / 16)
        return self.send_telegram("aA{}{}{}"
            .format(self.vdv_hex(address),
                self.vdv_hex(num_blocks),
                text.ljust(num_blocks*16)))
    
    def DS021a(self, address, stop_id, stop_text, change_text):
        """
        Line progress display texts
        
        address:
        The address of the display
        
        stop_id:
        The ID of the stop to send data for
        
        stop_text:
        The name of the stop
        
        change_text:
        Connection information
        """
        
        data = "\x03{:>02}\x04{}\x05{}".format(
            stop_id, stop_text, change_text)
        num_blocks, remainder = divmod(len(data), 16)
        return self.send_telegram("aL{}{}{}{}"
            .format(self.vdv_hex(address),
                self.vdv_hex(num_blocks),
                self.vdv_hex(remainder),
                data))

    def DS021t(self, address, text):
        """
        Destination text

        address:
        The address of the display

        text:
        The destination text
        """

        num_blocks = math.ceil(len(text) / 16)
        if '\n' not in text:
            text = text + '\n'
        return self.send_telegram("aA{}{}A0{}\n\n  "
            .format(self.vdv_hex(address),
                self.vdv_hex(num_blocks),
                text))
    
    def DS060(self, direction):
        """
        Query locating device status
        Reply: DS160
        
        direction:
        Either A, B or C
        """
        
        return self.parse_DS160(self.send_telegram("o{}S"
            .format(direction), reply_length = 3))
    
    def parse_DS160(self, telegram):
        if not telegram:
            return None
        status = telegram[2]
        statuses = {
            '0': 'ok_no_information',
            '1': 'information_read',
            '2': 'new_information'
        }
        reply = {
            'status': statuses.get(status, status)
        }
        return reply
    
    def DS601(self, direction):
        """
        Query locating device version
        Reply: DS1601
        
        direction:
        Either A, B or C
        """
        
        return self.parse_DS1201(self.send_telegram("o{}V"
            .format(direction), reply_length = 18))
    
    def parse_DS1601(self, telegram):
        if not telegram:
            return None
        version = int(telegram[2:])
        reply = {
            'version': version
        }
        return reply
    
    def DS061(self, direction):
        """
        Query locating device data
        Reply: DS161
        
        direction:
        Either A, B or C
        """
        
        return self.parse_DS161(self.send_telegram("o{}D"
            .format(direction), reply_length = 10))
    
    def parse_DS161(self, telegram):
        if not telegram:
            return None
        beacon_id_vdvhex = telegram[2:]
        beacon_id = 0x00
        for n in range(8):
            char = beacon_id_vdvhex[n]
            beacon_id |= self.vdv_hex(char) << ((7 - n) * 4)
        reply = {
            'beacon_id': beacon_id
        }
        return reply
    
    def DS068(self, channel, radio_telegram_type, delay, reporting_point_id, hand, line_number, course_number, destination_id, train_length):
        """
        Send an LSA radio telegram
        """
        
        radio_telegram = "{}{}{:02d}{}{}{:0>4}{}{:03d}{:02d}{:03d}{}".format(
            self.vdv_hex(channel),
            self.vdv_hex(0),
            radio_telegram_type,
            self.vdv_hex(delay),
            self.vdv_hex(6), # Number of extra bytes
            reporting_point_id,
            self.vdv_hex(hand),
            line_number,
            course_number,
            destination_id,
            self.vdv_hex(train_length)
        )
        return self.parse_DS160(self.send_telegram("oFM{}{}"
            .format(self.vdv_hex(math.ceil(len(radio_telegram) / 2)),
                radio_telegram), reply_length = 3))
    
    def GSP(self, address, line1 = "", line2 = ""):
        """
        Send GSP display data
        Reply: DS120
        
        address:
        The address of the display
        
        line1:
        The top line text
        
        line2:
        The bottom line text
        """
        
        lines = ""
        if line2:
            line1 += "\x0a" # Line feed (LF)
        lines += line1
        lines += line2
        lines += "\x0a\x0a"
        
        num_blocks = math.ceil(len(lines) / 16)
        remainder = len(lines) % 16
        if remainder:
            filler = " " * (16 - remainder)
        else:
            filler = ""
        
        data = "aA{address}{num_blocks}{lines}{filler}".format(
            address = self.vdv_hex(address),
            num_blocks = self.vdv_hex(num_blocks),
            lines = lines,
            filler = filler)
        
        return self.parse_DS120(self.send_telegram(data, reply_length = 2))